import tkinter as tk
from tkinter import *
from tkinter import ttk
from idlelib.tooltip import Hovertip
import threading

from eo_man import LOGGER

from ..controller.app_bus import AppBus, AppBusEventType
from ..controller.serial_controller import SerialController
from ..data.data_manager import DataManager

class SerialConnectionBar():

    def __init__(self, main: Tk, app_bus:AppBus, data_manager:DataManager, row:int):
        self.main = main
        self.app_bus = app_bus
        self.data_manager = data_manager
        self.serial_cntr = SerialController(app_bus)

        f = LabelFrame(main, text="Serial Connection", bd=1)#, relief=SUNKEN)
        f.grid(row=row, column=0, columnspan=1, sticky=W+E+N+S, pady=(0,2), padx=2)

        self.b_detect = ttk.Button(f, text="Detect Serial Ports")
        self.b_detect.pack(side=tk.LEFT, padx=(5, 5), pady=5)
        self.b_detect.config(command=lambda : self.detect_serial_ports_command(force_reload=True) )
        Hovertip(self.b_detect,"Serial port detection is sometime unstable. Please try again if device was not detected.",300)

        l = Label(f, text="Gateway Type: ")
        l.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.cb_device_type = ttk.Combobox(f, state="readonly", width="12") 
        self.cb_device_type.pack(side=tk.LEFT, padx=(0, 5), pady=5)
        self.cb_device_type['values'] = ['FAM14', 'FGW14-USB', 'FAM-USB', 'USB300']
        self.cb_device_type.set(self.cb_device_type['values'][0])
        self.cb_device_type.bind('<<ComboboxSelected>>', self.on_device_type_changed)

        l = Label(f, text="Serial Port: ")
        l.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.cb_serial_ports = ttk.Combobox(f, state="readonly", width="10") 
        self.cb_serial_ports.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.b_connect = ttk.Button(f, text="Connect", state=DISABLED, command=self.toggle_serial_connection_command)
        self.b_connect.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        s = ttk.Separator(f, orient=VERTICAL )
        s.pack(side=tk.LEFT, padx=(0,5), pady=0, fill="y")

        self.b_scan = ttk.Button(f, text="Scan for devices", state=DISABLED, command=self.scan_for_devices)
        self.b_scan.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.overwrite = tk.BooleanVar()
        self.cb_overwrite = ttk.Checkbutton(f, text="Overwrite existing values", variable=self.overwrite)
        self.cb_overwrite.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        s = ttk.Separator(f, orient=VERTICAL )
        s.pack(side=tk.LEFT, padx=(0,5), pady=0, fill="y")

        self.b_sync_ha_sender = ttk.Button(f, text="Write HA senders to devices", state=DISABLED, command=self.write_ha_senders_to_devices)
        Hovertip(self.b_sync_ha_sender,"Ensures sender configuration for Home Assistant is written into device memory.",300)
        self.b_sync_ha_sender.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.app_bus.add_event_handler(AppBusEventType.CONNECTION_STATUS_CHANGE, self.is_connected_handler)
        self.app_bus.add_event_handler(AppBusEventType.DEVICE_SCAN_STATUS, self.device_scan_status_handler)
        self.app_bus.add_event_handler(AppBusEventType.WRITE_SENDER_IDS_TO_DEVICES_STATUS, self.device_scan_status_handler)
        self.app_bus.add_event_handler(AppBusEventType.WINDOW_LOADED, self.on_window_loaded)
        

    def on_device_type_changed(self, event):
        self.cb_serial_ports['values'] = []
        self.cb_serial_ports.set('')

        self.detect_serial_ports_command()


    def write_ha_senders_to_devices(self):
        sender_list = {}
        for dev in self.data_manager.devices.values():
            sender_list[dev.external_id] = dev.additional_fields
        self.serial_cntr.write_sender_id_to_devices(45056, sender_list)
        

    def on_window_loaded(self, data):
        self.detect_serial_ports_command()


    def scan_for_devices(self):
        self.serial_cntr.scan_for_devices( self.overwrite.get() )


    def device_scan_status_handler(self, status:str):
        if status == 'FINISHED':
            self.is_connected_handler({'connected': self.serial_cntr.is_serial_connection_active()})
            self.main.config(cursor="")
            self.b_scan.config(state=NORMAL)
            self.b_connect.config(state=NORMAL)
            self.b_sync_ha_sender.config(state=NORMAL)
        if status == 'STARTED':
            self.b_scan.config(state=DISABLED)
            self.b_connect.config(state=DISABLED)
            self.main.config(cursor="watch")    #set cursor for waiting
            self.b_sync_ha_sender.config(state=DISABLED)


    def toggle_serial_connection_command(self):
        try:
            if not self.serial_cntr.is_serial_connection_active():
                self.b_detect.config(state=DISABLED)
                self.b_connect.config(state=DISABLED)
                self.b_scan.config(state=DISABLED)
                self.serial_cntr.establish_serial_connection(self.cb_serial_ports.get(), self.cb_device_type.get())
            else:
                self.serial_cntr.stop_serial_connection()
        except:
            # reset buttons
            self.is_connected_handler(data={'connected': self.serial_cntr.is_serial_connection_active()})
            LOGGER.exception("Was not able to detect serial ports.")


    def is_connected_handler(self, data:dict, skipp_serial_port_detection:bool=False):
        status = data.get('connected')
        
        self.main.config(cursor="")

        if status:
            self.b_connect.config(text="Disconnect", state=NORMAL)
            self.cb_serial_ports.config(state=DISABLED)
            self.b_detect.config(state=DISABLED)
            self.cb_device_type.config(state=DISABLED)
            
            if self.cb_device_type.get() == 'FAM14' and self.serial_cntr.is_fam14_connection_active():
                self.b_scan.config(state=NORMAL)
                self.b_sync_ha_sender.config(state=NORMAL)
            else:
                self.b_scan.config(state=DISABLED)
                self.b_sync_ha_sender.config(state=DISABLED)

        else:
            self.b_connect.config(text="Connect", state=NORMAL)
            self.b_detect.config(state=NORMAL)
            self.cb_serial_ports.config(state="readonly")
            self.cb_device_type.config(state="readonly")
            self.b_scan.config(state=DISABLED)
            self.b_sync_ha_sender.config(state=DISABLED)
            if not skipp_serial_port_detection: self.detect_serial_ports_command()


    def detect_serial_ports_command(self, force_reload:bool=False):

        def detect_serial_ports():
            try:
                self.main.config(cursor="watch")    #set cursor for waiting
                self.b_detect.config(state=DISABLED)
                self.cb_device_type.config(state=DISABLED)
                self.cb_serial_ports.config(state=DISABLED)
                serial_ports = self.serial_cntr.get_serial_ports(self.cb_device_type.get(), force_reload)
                self.cb_device_type.config(state=NORMAL)
                self.cb_device_type.config(state=NORMAL)
                self.cb_serial_ports['values'] = serial_ports
                if len(self.cb_serial_ports['values']) > 0:
                    self.cb_serial_ports.set(self.cb_serial_ports['values'][0])
                    self.b_connect.config(state=NORMAL)
                else:
                    self.b_connect.config(state=DISABLED)
                    self.cb_serial_ports.set('')
            except:
                # reset buttons
                LOGGER.exception("Was not able to detect serial ports.")
            else:
                self.is_connected_handler(data={'connected': False}, skipp_serial_port_detection=True)

        t = threading.Thread(target=detect_serial_ports)
        t.start()


import tkinter as tk
from tkinter import *
from tkinter import ttk
from tkinter import messagebox

from .checklistcombobox import ChecklistCombobox

from ..controller.app_bus import AppBus, AppBusEventType
from ..data.data_manager import DataManager
from ..data import data_helper
from ..data.const import CONF_EEP
from ..data.filter import DataFilter


class FilterBar():
    
    def __init__(self, main: Tk, app_bus:AppBus, data_manager:DataManager, row:int):
        self.app_bus = app_bus
        self.data_manager = data_manager
        
        f = LabelFrame(main, text= "Tabel Filter", bd=1, relief=SUNKEN)
        f.grid(row=row, column=0, columnspan=1, sticky=W+E+N+S, pady=(0,2), padx=2)
        self.root = f

        # filter name
        col = 0
        l = Label(f, text="Filter Name:")
        l.grid(row=0, column=col, padx=(3,3), sticky=W)

        self.cb_filtername = ttk.Combobox(f, width="14") 
        self.cb_filtername.grid(row=1, column=col, padx=(0,3) )
        self.cb_filtername.bind('<Return>', lambda e: [self.load_filter(), self.add_filter(False), self.apply_filter(e, True)] )

        col += 1
        self.btn_save_filter = ttk.Button(f, text="Load", command=self.load_filter)
        self.btn_save_filter.grid(row=1, column=col, padx=(0,3))

        col += 1
        self.btn_save_filter = ttk.Button(f, text="Remove", command=self.remove_filter)
        self.btn_save_filter.grid(row=0, column=col, padx=(0,3) )

        self.btn_save_filter = ttk.Button(f, text="Add", command=self.add_filter)
        self.btn_save_filter.grid(row=1, column=col, padx=(0,3), sticky=W+E )

        # global filter
        col += 1
        l = Label(f, text="Global Filter:")
        l.grid(row=0, column=col, padx=(0,3), sticky=W)

        self.global_filter = ttk.Entry(f, width="14") 
        self.global_filter.grid(row=1, column=col, padx=(0,3) )
        self.global_filter.bind('<Return>', self.apply_filter)
        self.global_filter.bind("<KeyRelease>", lambda e: self.cb_filtername.set('') )

        # address
        col += 1
        l = Label(f, text="Address:")
        l.grid(row=0, column=col, padx=(0,3), sticky=W)

        self.cb_device_address = ttk.Entry(f, width="14") 
        self.cb_device_address.grid(row=1, column=col, padx=(0,3) )
        self.cb_device_address.bind('<Return>', self.apply_filter)
        self.cb_device_address.bind("<KeyRelease>", lambda e: self.cb_filtername.set('') )

        # external address
        col += 1
        l = Label(f, text="External Address:")
        l.grid(row=0, column=col, padx=(0,3), sticky=W)

        self.cb_external_address = ttk.Entry(f, width="14") 
        self.cb_external_address.grid(row=1, column=col, padx=(0,3) )
        self.cb_external_address.bind('<Return>', self.apply_filter)
        self.cb_external_address.bind("<KeyRelease>", lambda e: self.cb_filtername.set('') )

        # device type
        col += 1
        l = Label(f, text="Device Type:")
        l.grid(row=0, column=col, padx=(0,3), sticky=W)

        values = sorted(set([info['hw-type'] for info in data_helper.EEP_MAPPING]))
        self.cb_device_type = ChecklistCombobox(f, values=values, width="14") 
        self.cb_device_type.grid(row=1, column=col, padx=(0,3) )
        self.cb_device_type.bind('<Return>', self.apply_filter)
        self.cb_device_type.bind("<KeyRelease>", lambda e: self.cb_filtername.set('') )

        # eep
        col += 1
        l = Label(f, text="Device EEP:")
        l.grid(row=0, column=col, padx=(0,3), sticky=W)

        values = data_helper.get_all_eep_names()
        self.cb_device_eep = ChecklistCombobox(f, values=values, width="14") 
        self.cb_device_eep.grid(row=1, column=col, padx=(0,3))
        self.cb_device_eep.bind('<Return>', self.apply_filter)
        self.cb_device_eep.bind("<KeyRelease>", lambda e: self.cb_filtername.set('') )

        # used in ha
        col += 1
        self.var_in_ha = tk.IntVar()
        self.var_in_ha.set(1)
        self.cb_in_ha = ttk.Checkbutton(f, text="Used in HA", variable=self.var_in_ha)
        self.cb_in_ha.grid(row=1, column=col, padx=(0,3) )
        self.cb_in_ha.bind('<Return>', self.apply_filter)
        self.cb_in_ha.bind('<Button-1>', lambda e: self.cb_filtername.set('') )


        # button reset
        col += 1
        self.btn_clear_filter = ttk.Button(f, text="Reset", command=self.reset_filter)
        self.btn_clear_filter.grid(row=1, column=col, padx=(0,3) )

        col += 1
        self.btn_apply_filter = ttk.Button(f, text="Apply", command=self.apply_filter)
        self.btn_apply_filter.grid(row=1, column=col, padx=(0,3) )

        self.app_bus.add_event_handler(AppBusEventType.SET_DATA_TABLE_FILTER, self.on_set_filter_handler)
        self.app_bus.add_event_handler(AppBusEventType.ADDED_DATA_TABLE_FILTER, self.on_filter_added_handler)
        self.app_bus.add_event_handler(AppBusEventType.REMOVED_DATA_TABLE_FILTER, self.on_filter_removed_handler)

        ## initiall add all filters
        for d in self.data_manager.data_fitlers.values():
            self.on_filter_added_handler(d)
        selected_fn = self.data_manager.selected_data_filter_name
        if selected_fn is not None and len(selected_fn) > 0 and selected_fn in self.data_manager.data_fitlers.keys():
            self.on_set_filter_handler(self.data_manager.data_fitlers[selected_fn])
            self.apply_filter()



    def on_set_filter_handler(self, filter:DataFilter):
        self.set_widget_values(filter)

    def on_filter_added_handler(self, filter:DataFilter):
        # always take entire list to be in sync
        self.cb_filtername['values'] = sorted([k for k in self.data_manager.data_fitlers.keys()])

    def on_filter_removed_handler(self, filter:DataFilter):
        # always take entire list to be in sync
        self.cb_filtername['values'] = sorted([k for k in self.data_manager.data_fitlers.keys()])
        self.reset_filter()

    def add_filter(self, show_error_message:bool=True):
        filter_obj = self.get_filter_object()
        if filter_obj and len(filter_obj.name) > 1:
            self.data_manager.add_filter(filter_obj)
            
        elif show_error_message:
            messagebox.showerror(title="Error: Cannot add filter", message="Please, provdied a valid filter name!")

    def remove_filter(self):
        filter_obj = self.get_filter_object()
        if len(filter_obj.name) > 1:
            self.data_manager.remove_filter(filter_obj)
        else:
            messagebox.showerror(title="Error: Cannot remove filter", message="Please, provdied a valid filter name!")

    def set_widget_values(self, filter:DataFilter):
        if filter is not None:
            self.cb_filtername.set(filter.name)
            self.global_filter.delete(0, END)
            self.global_filter.insert(END, ', '.join(filter.global_filter))
            self.cb_device_address.delete(0, END)
            self.cb_device_address.insert(END, ', '.join(filter.device_address_filter))
            self.cb_external_address.delete(0, END)
            self.cb_external_address.insert(END, ', '.join(filter.device_external_address_filter))
            self.select_ChecklistCombobox(self.cb_device_type, filter.device_type_filter)
            self.select_ChecklistCombobox(self.cb_device_eep, filter.device_eep_filter)
        else:
            self.cb_filtername.set('')
            self.global_filter.delete(0, END)
            
            for cb in [self.cb_device_address, self.cb_external_address, self.cb_device_eep, self.cb_device_type]:
                cb.delete(0, END)
                if isinstance(cb, ChecklistCombobox):
                    for var in cb.variables:
                        var.set(0)

    def load_filter(self):
        filter_name = self.cb_filtername.get()
        if filter_name in self.data_manager.data_fitlers.keys():
            df:DataFilter = self.data_manager.data_fitlers[filter_name]
            self.set_widget_values(df)

            # load shall only present the values and not apply the filter
            # self.apply_filter()


    def select_ChecklistCombobox(self, widget:ChecklistCombobox, values:list[str]):
        widget.set(', '.join(values))
        for i in range(0,len(widget.checkbuttons)):
            widget.variables[i].set( 1 if widget.checkbuttons[i].cget('text') in values else 0 )

    def apply_filter(self, event=None, reset_for_no_filter_name:bool=False):
        filter = self.get_filter_object()
        # trigger reset if filter name is empty and if enter was pushed on filter name field
        if reset_for_no_filter_name:
            if filter is None or filter.name is None or len(filter.name) == 0:
                filter = None

        self.app_bus.fire_event(AppBusEventType.SET_DATA_TABLE_FILTER, filter)


    def reset_filter(self):
        self.app_bus.fire_event(AppBusEventType.SET_DATA_TABLE_FILTER, None)

    def get_str_array(self, widget:Widget) -> list[str]:
        widget_value = ''
        try:
            widget_value = widget.get()
        except:
            pass
        if isinstance(widget_value, str):
            return [g.strip() for g in widget_value.replace(';',',').split(',') if len(g.strip()) > 0 ]
        elif isinstance(widget_value, list):
            return [g.strip() for g in widget_value if len(g.strip()) > 0 ]
        else:
            return []


    def get_filter_object(self):

        filter_obj = DataFilter(
            name = self.cb_filtername.get(),
            global_filter=self.get_str_array(self.global_filter),
            device_address_filter=self.get_str_array(self.cb_device_address),
            device_external_address_filter=self.get_str_array(self.cb_external_address),
            device_type_filter=self.get_str_array(self.cb_device_type),
            device_eep_filter=self.get_str_array(self.cb_device_eep),
        )
        
        reset  = len(filter_obj.global_filter) == 0 
        reset &= len(filter_obj.device_address_filter) == 0 
        reset &= len(filter_obj.device_external_address_filter) == 0 
        reset &= len(filter_obj.device_type_filter) == 0 
        reset &= len(filter_obj.device_eep_filter) == 0 

        if reset: 
            return None
        else: 
            return filter_obj
        
        


import sys
import os
import argparse
from typing import Final
from logging.handlers import RotatingFileHandler

PACKAGE_NAME: Final = 'eo_man'

# load same path like calling the app via 'python -m eo-man'
PROJECT_DIR = os.path.dirname(__file__)
file_dir = os.path.join( PROJECT_DIR, '..')
sys.path.append(file_dir)
__import__(PACKAGE_NAME)
__package__ = PACKAGE_NAME

from eo_man import load_dep_homeassistant, LOGGER
load_dep_homeassistant()

from .data.app_info import ApplicationInfo
from .data.data_manager import DataManager
from .data.ha_config_generator import HomeAssistantConfigurationGenerator
from .view.main_panel import MainPanel
from .controller.app_bus import AppBus, AppBusEventType

import logging


def cli_argument():
    p = argparse.ArgumentParser(
        description=
"""EnOcean Device Manager (https://github.com/grimmpp/enocean-device-manager) allows you to managed your EnOcean devices and to generate 
Home Assistant Configurations for the Home Assistant Eltako Integration (https://github.com/grimmpp/home-assistant-eltako).""")
    p.add_argument('-v', '--verbose', help="Logs all messages.", action='store_true')
    p.add_argument('-c', "--app_config", help="Filename of stored application configuration. Filename must end with '.eodm'.", default=None)
    p.add_argument('-ha', "--ha_config", help="Filename for Home Assistant Configuration for Eltako Integration. By passing the filename it will disable the GUI and only generate the Home Assistant Configuration file.")
    return p.parse_args()


def init_logger(app_bus:AppBus, log_level:int=logging.INFO):
    file_handler = RotatingFileHandler(os.path.join(PROJECT_DIR, "enocean-device-manager.log"),
                                       mode='a', maxBytes=10*1024*1024, backupCount=2, encoding=None, delay=0)

    logging.basicConfig(format='%(asctime)s %(name)s %(levelname)s %(message)s ', #'%(message)s ', 
                        level=logging.INFO,
                        handlers=[ file_handler, logging.StreamHandler() ])
    global LOGGER
    LOGGER = logging.getLogger(PACKAGE_NAME)
    LOGGER.setLevel(log_level)
    file_handler.setLevel(logging.DEBUG)
    LOGGER.info("Start Application eo_man")
    LOGGER.info(ApplicationInfo.get_app_info_as_str())
    # add print log messages for log message view on command line as debug
    def print_log_event(e:dict):
        log_level = e.get('log-level', 'INFO')
        log_level_int = logging.getLevelName(log_level)
        if log_level_int >= LOGGER.getEffectiveLevel():
            LOGGER.log(log_level_int, str(e['msg']))
    app_bus.add_event_handler(AppBusEventType.LOG_MESSAGE, print_log_event)


def main():
    opts = cli_argument()

    if hasattr(opts, 'help') and opts.help:
        return
    
    # init application message BUS
    app_bus = AppBus()

    init_logger(app_bus, logging.DEBUG if opts.verbose else logging.INFO)

    # init DATA MANAGER
    data_manager = DataManager(app_bus)

    # initially load from file application data
    if opts.app_config and opts.app_config.endswith('.eodm'):
       e = {'msg': f"Initially load data from file {opts.app_config}", 'color': 'darkred'}
       app_bus.fire_event(AppBusEventType.LOG_MESSAGE, e)
       data_manager.load_application_data_from_file(opts.app_config)
    elif opts.app_config:
       e = {'msg': f"Invalid filename {opts.app_config}. It must end with '.eodm'", 'color': 'darkred'}
       app_bus.fire_event(AppBusEventType.LOG_MESSAGE, e)

    # generate home assistant config instead of starting GUI
    if opts.app_config and opts.app_config.endswith('.eodm') and opts.ha_config:
        HomeAssistantConfigurationGenerator(app_bus, data_manager).save_as_yaml_to_file(opts.ha_config)
    else:
        MainPanel(app_bus, data_manager)

if __name__ == "__main__":
    main()



from eltakobus.device import SensorInfo, KeyFunction
from eltakobus.util import b2s, AddressExpression
from eltakobus.device import BusObject, FAM14
from eltakobus.message import *
from eltakobus.eep import *

from homeassistant.const import CONF_ID, CONF_DEVICE, CONF_DEVICES, CONF_NAME, CONF_PLATFORM, CONF_TYPE, CONF_DEVICE_CLASS, CONF_TEMPERATURE_UNIT, UnitOfTemperature, Platform

from .data_helper import *
from .const import *

class Device():
    """Data representation of a device"""
    static_info:dict={}
    address:str = None
    channel:int=None
    dev_size:int=None
    external_id:str=None
    device_type:str=None
    key_function:str=None
    version:str=None
    name:str=None
    comment:str = ""
    base_id:str=None
    bus_device:bool=False
    memory_entries:list[SensorInfo]=[]  # only used for bus devices

    # vars for ha
    use_in_ha:bool=False
    ha_platform:Platform=None
    eep:str=None
    additional_fields:dict={}


    def __init__(self, 
                 address:str=None, 
                 bus_device:bool=False,
                 dev_size:int=1,
                 channel:int=1, 
                 external_id:str=None, 
                 device_type:str=None, 
                 version:str=None, 
                 name:str=None, 
                 comment:str=None, 
                 base_id:str=None, 
                 use_in_ha:bool=False,
                 memory_entries:list[SensorInfo]=[]):
        
        self.address = address
        self.bus_device = bus_device
        self.channel = channel
        self.dev_size = dev_size
        self.external_id = external_id
        self.device_type = device_type
        self.version = version
        self.name = name
        self.comment = comment
        self.base_id = base_id
        self.use_in_ha = use_in_ha
        self.memory_entries = memory_entries

    def is_fam14(self) -> bool:
        return self.device_type is not None and FAM14.__name__ in self.device_type

    def is_fam_usb(self) -> bool:
        return self.device_type is not None and ('FAM-USB' in self.device_type or 'FAM_USB' in self.device_type)
    
    def is_fgw14_usb(self) -> bool:
        return self.device_type is not None and 'FGW14_USB' in self.device_type

    def is_usb300(self) -> bool:
        return self.device_type is not None and 'USB300' in self.device_type

    def is_gateway(self) -> bool:
        return self.is_wired_gateway() or self.is_wireless_tranceiver()
    
    def is_wired_gateway(self) -> bool:
        return self.is_fam14() or self.is_fgw14_usb()

    def is_wireless_tranceiver(self) -> bool:
        return self.is_usb300() or self.is_fam_usb() or (self.device_type is not None and 'Wireless Tranceiver' in self.device_type)
    
    def is_bus_device(self) -> bool:
        return self.bus_device
        # return self.address.startswith('00-00-00-')            

    @classmethod
    def merge_devices(cls, device1, device2):
        d1:Device = device1
        d2:Device = device2

        if d1.external_id != d2.external_id: return None

        d1.address = d2.address
        d1.channel = d2.channel
        d1.dev_size = d2.dev_size
        d1.device_type = d2.device_type
        d1.version = d2.version
        if d1.name == 'unknown': d1.name = d2.name
        if d1.comment is None or d1.comment == '': d1.comment = d2.comment
        d1.base_id = d2.base_id
        d1.memory_entries = d2.memory_entries
        if not d1.bus_device: d1.bus_device = d2.bus_device
        if d1.key_function is None or d1.key_function == '': d1.key_function = d2.key_function
        d1.use_in_ha = d2.use_in_ha
        d1.eep = d2.eep

        for k,v in d2.additional_fields.items():
            if k not in d1.additional_fields:
                d1.additional_fields[k] = v

    @classmethod
    async def async_get_bus_device_by_natvice_bus_object(cls, device: BusObject, fam14: FAM14, channel:int=1):
        bd = Device()
        bd.additional_fields = {}
        id = device.address + channel -1
        bd.address = a2s( id )
        bd.channel = channel
        bd.dev_size = device.size
        bd.base_id = await fam14.get_base_id()
        bd.device_type = type(device).__name__
        bd.version = '.'.join(map(str,device.version))
        bd.key_function = ''
        bd.comment = ''
        bd.bus_device = True
        if isinstance(device, FAM14):
            bd.external_id = bd.base_id
        else:
            bd.external_id = a2s( (await fam14.get_base_id_in_int()) + id )
        bd.memory_entries = [m for m in (await device.get_all_sensors()) if b2s(m.dev_adr) == bd.address]
        # print(f"{bd.device_type} {bd.address}")
        # print_memory_entires( bd.memory_entries)
        # print("\n")
        bd.name = f"{bd.device_type} {bd.address}"
        if bd.is_fam14():
            bd.name = f"{bd.device_type} {bd.external_id}"
        
        if bd.dev_size > 1:
            bd.name += f" ({bd.channel}/{bd.dev_size})"

        Device.set_suggest_ha_config(bd)

        if bd.device_type == BusObject.__name__:
            bd.device_type = "unknown"
            bd.name = f"unknown device  {bd.address}"

        return bd
    
    @classmethod
    def set_suggest_ha_config(cls, device):
        id = int.from_bytes( AddressExpression.parse(device.address)[0], 'big')
        info:dict = find_device_info_by_device_type(device.device_type)
        if info is not None:
            device.use_in_ha = device.device_type != BusObject.__name__
            device.ha_platform = info.get(CONF_TYPE, None)
            device.eep = info.get(CONF_EEP, None)

            if info.get('sender_eep', None):
                device.additional_fields['sender'] = {
                    CONF_ID: a2s( SENDER_BASE_ID + id ),
                    CONF_EEP: info.get('sender_eep')
                }

            if info.get(CONF_TYPE, None) == Platform.COVER:
                device.additional_fields[CONF_DEVICE_CLASS] = 'shutter'
                device.additional_fields[CONF_TIME_CLOSES] = 25
                device.additional_fields[CONF_TIME_OPENS] = 25

            if info.get(CONF_TYPE, None) == Platform.CLIMATE:
                device.additional_fields[CONF_TEMPERATURE_UNIT] = f"'{UnitOfTemperature.KELVIN}'"
                device.additional_fields[CONF_MIN_TARGET_TEMPERATURE] = 16
                device.additional_fields[CONF_MAX_TARGET_TEMPERATURE] = 25
                thermostat = BusObjectHelper.find_sensor(device.memory_entries, device.address, device.channel, in_func_group=1)
                if thermostat:
                    device.additional_fields[CONF_ROOM_THERMOSTAT] = {}
                    device.additional_fields[CONF_ROOM_THERMOSTAT][CONF_ID] = b2s(thermostat.sensor_id)
                    device.additional_fields[CONF_ROOM_THERMOSTAT][CONF_EEP] = A5_10_06.eep_string   #TODO: derive EEP from switch/sensor function
                #TODO: cooling_mode
    
    @classmethod
    def get_decentralized_device_by_telegram(cls, msg: RPSMessage):
        bd = Device()
        bd.address = b2s( msg.address )
        bd.bus_device = False
        bd.base_id = '00-00-00-00'
        bd.device_type = 'unknown'
        bd.version = 'unknown'
        bd.comment = ''
        if int.from_bytes( msg.address, "big") > 0x0000FFFF:
            bd.external_id = b2s( msg.address )
        else:
            bd.external_id = 'unknown'
        bd.name = 'unknown'
        return bd
    
    @classmethod
    def get_centralized_device_by_telegram(cls, msg: RPSMessage, base_id:str, external_id:str):
        bd = Device()
        bd.bus_device = True
        bd.address = b2s( msg.address )
        bd.base_id = base_id
        bd.external_id = external_id
        bd.device_type = 'unknown'
        bd.version = 'unknown'
        bd.comment = ''
        bd.name = 'unknown'
        return bd
    
    @classmethod
    async def async_get_decentralized_device_by_sensor_info(cls, sensor_info:SensorInfo, device: BusObject, fam14: FAM14, channel:int=1):
        return cls.get_decentralized_device_by_sensor_info(sensor_info, (await fam14.get_base_id()) )
    
    @classmethod 
    def get_decentralized_device_by_sensor_info(cls, sensor_info:SensorInfo, base_id:str='00-00-00-00'):
        bd = Device()
        bd.address = sensor_info.sensor_id_str
        bd.comment = ''
        bd.key_function = KeyFunction(sensor_info.key_func).name
        bd.eep = get_eep_from_key_function_name(sensor_info.key_func)
        bd.name = f"{get_name_from_key_function_name(sensor_info.key_func)} {sensor_info.sensor_id_str}"
        # found sensor by EEP in KeyFunction
        if bd.eep and bd.name:
            bd.use_in_ha = True
            bd.device_type = 'Sensor'
            bd.ha_platform = Platform.SENSOR
        # buttons but ignore virtual buttons from HA
        elif 'PUSH_BUTTON' in KeyFunction(sensor_info.key_func).name and not sensor_info.sensor_id_str.startswith('00-00-'):
            bd.use_in_ha = True
            bd.device_type = 'Button'
            bd.ha_platform = Platform.BINARY_SENSOR
            bd.eep = F6_02_01.eep_string
            bd.name = f"Button {sensor_info.sensor_id_str}"
        # is from FTS14EM
        elif int.from_bytes(sensor_info.sensor_id, "big") < 0x1500:
            bd.use_in_ha = True
            bd.device_type = 'FTS14EM Button'
            bd.ha_platform = Platform.BINARY_SENSOR
            bd.eep = F6_02_01.eep_string
            bd.name = 'FTS14EM Button ' + sensor_info.sensor_id_str
        elif bd.address.startswith('00-00-B') or 'FROM_CONTROLLER' in bd.key_function:
            bd.use_in_ha = False
            bd.device_type = 'Smart Home'
            bd.ha_platform = ''
            bd.eep = ''
            bd.name = 'HA Contoller ' + sensor_info.sensor_id_str
        elif 'WEATHER_STATION' in bd.key_function:
            bd.use_in_ha = True
            bd.device_type = 'Weather Station'
            bd.ha_platform = Platform.SENSOR
            bd.eep = 'A5-04-02'
            bd.name = 'Weather Station ' + sensor_info.sensor_id_str

        if sensor_info.sensor_id_str.startswith('00-00-'):
            bd.external_id = a2s( int.from_bytes(sensor_info.sensor_id, "big") + int.from_bytes(AddressExpression.parse(base_id)[0], 'big') )
            bd.base_id = base_id
        else:
            bd.external_id = sensor_info.sensor_id_str
            bd.base_id = '00-00-00-00'
        bd.version = 'unknown'
        return bd



import os
from termcolor import colored

from .const import *
from homeassistant.const import CONF_ID, CONF_DEVICE, CONF_DEVICES, CONF_NAME, CONF_PLATFORM, CONF_TYPE, CONF_DEVICE_CLASS, CONF_TEMPERATURE_UNIT, UnitOfTemperature, Platform

from eltakobus.device import BusObject, FAM14, SensorInfo, KeyFunction
from eltakobus.message import *
from eltakobus.eep import *
from eltakobus.device import SensorInfo, KeyFunction
from eltakobus.util import b2s, AddressExpression

SENDER_BASE_ID = 0x0000B000

EEP_MAPPING = [
    {'hw-type': 'BusObject', CONF_EEP: 'unknown', CONF_TYPE: 'unknown', 'description': 'unknown bus device'},
    {'hw-type': 'FAM14', 'description': 'Bus Gateway'},
    {'hw-type': 'FGW14_USB', 'description': 'Bus Gateway'},

    {'hw-type': 'FTS14EM', CONF_EEP: 'F6-02-01', CONF_TYPE: Platform.BINARY_SENSOR, 'description': 'Rocker switch', 'address_count': 1},
    {'hw-type': 'FTS14EM', CONF_EEP: 'F6-02-02', CONF_TYPE: Platform.BINARY_SENSOR, 'description': 'Rocker switch', 'address_count': 1},
    {'hw-type': 'FTS14EM', CONF_EEP: 'F6-10-00', CONF_TYPE: Platform.BINARY_SENSOR, 'description': 'Window handle', 'address_count': 1},
    {'hw-type': 'FTS14EM', CONF_EEP: 'D5-00-01', CONF_TYPE: Platform.BINARY_SENSOR, 'description': 'Contact sensor', 'address_count': 1},
    {'hw-type': 'FTS14EM', CONF_EEP: 'A5-08-01', CONF_TYPE: Platform.BINARY_SENSOR, 'description': 'Occupancy sensor', 'address_count': 1},

    # {'hw-type': 'FWG14', CONF_EEP: 'A5-13-01', CONF_TYPE: Platform.SENSOR, 'description': 'Weather station', 'address_count': 1},
    {'hw-type': 'FTS14EM', CONF_EEP: 'A5-12-01', CONF_TYPE: Platform.SENSOR, 'description': 'Window handle', 'address_count': 1},
    {'hw-type': 'FSDG14', CONF_EEP: 'A5-12-02', CONF_TYPE: Platform.SENSOR, 'description': 'Automated meter reading - electricity', 'address_count': 1},
    {'hw-type': 'F3Z14D', CONF_EEP: 'A5-13-01', CONF_TYPE: Platform.SENSOR, 'description': 'Automated meter reading - gas', 'address_count': 1},
    {'hw-type': 'F3Z14D', CONF_EEP: 'A5-12-03', CONF_TYPE: Platform.SENSOR, 'description': 'Automated meter reading - water', 'address_count': 1},

    {'hw-type': 'FUD14', CONF_EEP: 'A5-38-08', 'sender_eep': 'A5-38-08', CONF_TYPE: Platform.LIGHT, 'PCT14-function-group': 3, 'PCT14-key-function': 32, 'description': 'Central command - gateway', 'address_count': 1},
    {'hw-type': 'FUD14_800W', CONF_EEP: 'A5-38-08', 'sender_eep': 'A5-38-08', CONF_TYPE: Platform.LIGHT, 'PCT14-function-group': 3, 'PCT14-key-function': 32, 'description': 'Central command - gateway', 'address_count': 1},

    {'hw-type': 'FMZ14', CONF_EEP: 'M5-38-08', 'sender_eep': 'A5-38-08', CONF_TYPE: Platform.LIGHT, 'PCT14-function-group': 1, 'description': 'Eltako relay', 'address_count': 1},
    {'hw-type': 'FSR14', CONF_EEP: 'M5-38-08', 'sender_eep': 'A5-38-08', CONF_TYPE: Platform.LIGHT, 'PCT14-function-group': 2, 'PCT14-key-function': 51, 'description': 'Eltako relay', 'address_count': 1},
    {'hw-type': 'FSR14_1x', CONF_EEP: 'M5-38-08', 'sender_eep': 'A5-38-08', CONF_TYPE: Platform.LIGHT, 'PCT14-function-group': 2, 'PCT14-key-function': 51, 'description': 'Eltako relay', 'address_count': 1},
    {'hw-type': 'FSR14_x2', CONF_EEP: 'M5-38-08', 'sender_eep': 'A5-38-08', CONF_TYPE: Platform.LIGHT, 'PCT14-function-group': 2, 'PCT14-key-function': 51, 'description': 'Eltako relay', 'address_count': 2},
    {'hw-type': 'FSR14_4x', CONF_EEP: 'M5-38-08', 'sender_eep': 'A5-38-08', CONF_TYPE: Platform.LIGHT, 'PCT14-function-group': 2, 'PCT14-key-function': 51, 'description': 'Eltako relay', 'address_count': 4},
    {'hw-type': 'F4SR14_LED', CONF_EEP: 'M5-38-08', 'sender_eep': 'A5-38-08', CONF_TYPE: Platform.LIGHT, 'PCT14-function-group': 2, 'PCT14-key-function': 51, 'description': 'Eltako relay', 'address_count': 4},

    {'hw-type': 'FSB14', CONF_EEP: 'G5-3F-7F', 'sender_eep': 'H5-3F-7F', CONF_TYPE: Platform.COVER, 'PCT14-function-group': 2, 'PCT14-key-function': 31, 'description': 'Eltako cover', 'address_count': 2},

    {'hw-type': 'FAE14SSR', CONF_EEP: 'A5-10-06', 'sender_eep': 'A5-10-06', CONF_TYPE: Platform.CLIMATE, 'PCT14-function-group': 3, 'PCT14-key-function': 65, 'description': 'Eltako heating/cooling', 'address_count': 2},
]

ORG_MAPPING = {
    5: {'Telegram': 'RPS', 'RORG': 'F6', CONF_NAME: 'Switch', CONF_TYPE: Platform.BINARY_SENSOR, CONF_EEP: 'F6-02-01' },
    6: {'Telegram': '1BS', 'RORG': 'D5', CONF_NAME: '1 Byte Communication', CONF_TYPE: Platform.SENSOR, CONF_EEP: 'D5-??-??' },
    7: {'Telegram': '4BS', 'RORG': 'A5', CONF_NAME: '4 Byte Communication', CONF_TYPE: Platform.SENSOR, CONF_EEP: 'A5-??-??' },
}

SENSOR_MESSAGE_TYPES = [EltakoWrappedRPS, EltakoWrapped4BS, RPSMessage, Regular4BSMessage, Regular1BSMessage, EltakoMessage]


def get_all_eep_names():
    subclasses = set()
    work = [EEP]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return sorted(set([s.__name__.upper() for s in subclasses if len(s.__name__) == 8 and s.__name__.count('_') == 2]))

def find_eep_by_name(eep_name:str) -> EEP:
    for child in EEP.__subclasses__():
        if child.__name__.replace('_', '-').upper() == eep_name.replace('_', '-').upper():
            return child
        for sub_child in child.__subclasses__():
            if sub_child.eep_string.replace('_', '-').upper() == eep_name.replace('_', '-').upper():
                return sub_child
    return None

def get_values_for_eep(eep:EEP, message:EltakoMessage) -> list[str]:
    properties_as_str = []
    for k, v in eep.decode_message(message).__dict__.items():
        properties_as_str.append(f"{str(k)[1:] if str(k).startswith('_') else str(k)}: {str(v)}")

    return properties_as_str

def a2s(address:int, length:int=4):
    """address to string"""
    if address is None:
        return ""
    
    return b2s( address.to_bytes(length, byteorder = 'big') )

def find_device_info_by_device_type(device_type:str) -> dict:
    for i in EEP_MAPPING:
        if i['hw-type'] == device_type:
            return i
    return None

def get_eep_from_key_function_name(kf: KeyFunction) -> str:
    pos = KeyFunction(kf).name.find('EEP_')
    if pos > -1:
        substr = KeyFunction(kf).name[pos+4:pos+4+8].replace('_', '-')
        return substr
    return None

def get_name_from_key_function_name(kf: KeyFunction) -> str:
    pos = KeyFunction(kf).name.find('_ACCORDING_')
    if pos > -1:
        substr = KeyFunction(kf).name[0:pos].replace('_', ' ').lower().title()
        return substr
    return ""

def a2i(address:str):
    return int.from_bytes(AddressExpression.parse(address)[0], 'big')

class BusObjectHelper():

    @classmethod
    def find_sensors(cls, sensors:list[SensorInfo], dev_id: int, channel:int, in_func_group: int) -> list[SensorInfo]:
        result = []
        for s in sensors: 
            if s.dev_id == dev_id and s.channel == channel and s.in_func_group == in_func_group:
                result.append(s)
        return result

    @classmethod
    def find_sensor(cls, sensors:list[SensorInfo], dev_id: int, channel:int, in_func_group: int) -> SensorInfo:
        l = cls.find_sensors(sensors, dev_id, channel, in_func_group)
        if len(l) > 0:
            return l[0]
        return None
    
def add_addresses(adr1:str, adr2:str) -> str:
    _adr1 = int.from_bytes( AddressExpression.parse(adr1)[0], 'big')
    _adr2 = int.from_bytes( AddressExpression.parse(adr2)[0], 'big')
    return a2s(_adr1 + _adr2)
    
def print_memory_entires(sensors: list[SensorInfo]) -> None:
    for _s in sensors:
        s:SensorInfo = _s
        print(f"{s.memory_line}: {b2s(s.sensor_id, ' ')} {hex(s.key)} {hex(s.key_func)} {hex(s.channel)} (FG: {s.in_func_group})")
        


import os
from setuptools import setup, find_packages
from distutils.core import setup



base_dir = os.path.dirname(__file__)

# with open(os.path.join(base_dir, 'README.md'), encoding="utf-8") as f:
with open('README.md', encoding="utf-8") as f:
    long_description = f.read()

# with open(os.path.join(base_dir, 'LICENSE'), encoding="utf-8") as f:
with open('LICENSE', encoding="utf-8") as f:
    license = f.read()

required = ['eltako14bus==0.0.46', 'requests==2.31.0', 'enocean==0.60.1', 'pyserial', 'pyserial-asyncio', 'aiocoap', 
            # 'homeassistant', 
            'pyyaml', 
            'termcolor', 'strenum', 'pillow', 'numpy',
            # 'tzlocal', 
            'tkinterhtml']



setup(
    name='eo_man',
    version='0.1.11',
    package_dir={'eo_man':"eo_man"},
    # packages=find_packages("./eo-man"),
    #package_data={'': ['*.png']},
    package_data={'': ['icons/*']},
    package=["eo_man"],
    include_package_data=True,
    install_requires=required,
    author="Philipp Grimm",
    description="Tool to managed EnOcean Devices and to generate Home Assistant Configuration.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license=license,
    url="https://github.com/grimmpp/enocean-device-manager",
    python_requires='>=3.7',
)

#python setup.py bdist_wheel
